from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import (
    PrefixesQuery,
    ResultsQuery,
    prefixes_table,
    results_table,
)
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetInvalidPrefixes(PrefixesQuery):
    """
    Return the prefixes with unexpected behavior
    (see `GetPrefixesWithAmplification` and `GetPrefixesWithLoops`.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetInvalidPrefixes
        >>> rows = GetInvalidPrefixes().execute(client, "test_invalid_prefixes")
        >>> [x["probe_dst_prefix"] for x in rows]
        ['::ffff:201.0.0.0', '::ffff:202.0.0.0']
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT probe_dst_prefix
        FROM {prefixes_table(measurement_id)}
        WHERE {self.filters(subset)} AND (has_amplification OR has_loops)
        """


@dataclass(frozen=True)
class GetPrefixesWithAmplification(ResultsQuery):
    """
    Return the prefixes for which we have more than one reply per (flow ID, TTL).

    Important:
        This query assumes that a single probe is sent per (flow ID, TTL) pair.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetPrefixesWithAmplification
        >>> rows = GetPrefixesWithAmplification().execute(client, "test_invalid_prefixes")
        >>> [x["probe_dst_prefix"] for x in rows]
        ['::ffff:202.0.0.0']
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT DISTINCT
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            -- This column is to simplify the InsertPrefixes query.
            1 AS has_amplification
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port,
            probe_ttl
        )
        HAVING count() > 1
        """


@dataclass(frozen=True)
class GetPrefixesWithLoops(ResultsQuery):
    """
    Return the prefixes for which an IP address appears multiple time for a single flow ID.

    Important:
        Prefixes with amplification (multiple replies per probe) may trigger a false positive
        for this query, since we do not check that the IP appears at two *different* TTLs.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetPrefixesWithLoops
        >>> GetPrefixesWithLoops().execute(client, "test_invalid_prefixes")
        [{'probe_protocol': 1, 'probe_src_addr': '::ffff:100.0.0.1', 'probe_dst_prefix': '::ffff:201.0.0.0', 'has_loops': 1}]
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT DISTINCT
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            -- This column is to simplify the InsertPrefixes query.
            1 AS has_loops
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port
        )
        HAVING uniqExact(reply_src_addr) < count()
        """



from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import ip_in
from diamond_miner.queries.query import ProbesQuery, probes_table
from diamond_miner.typing import IPNetwork


class GetProbes(ProbesQuery):
    """
    Return the cumulative number of probes sent at a specified round.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetProbes
        >>> row = GetProbes(round_eq=1).execute(client, 'test_nsdi_example')[0]
        >>> row["probe_protocol"]
        1
        >>> row["probe_dst_prefix"]
        '::ffff:200.0.0.0'
        >>> sorted(row["probes_per_ttl"])
        [[1, 6], [2, 6], [3, 6], [4, 6]]
        >>> row = GetProbes(round_eq=2).execute(client, 'test_nsdi_example')[0]
        >>> sorted(row["probes_per_ttl"])
        [[1, 11], [2, 18], [3, 18], [4, 18]]
        >>> row = GetProbes(round_eq=3).execute(client, 'test_nsdi_example')[0]
        >>> sorted(row["probes_per_ttl"])
        [[1, 11], [2, 20], [3, 27], [4, 27]]
        >>> row = GetProbes(round_eq=3, probe_ttl_geq=2, probe_ttl_leq=3).execute(client, 'test_nsdi_example')[0]
        >>> sorted(row["probes_per_ttl"])
        [[2, 20], [3, 27]]
        >>> GetProbes(round_eq=4).execute(client, 'test_nsdi_example')
        []
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq, "`round_eq` must be specified."
        return f"""
        SELECT
            probe_protocol,
            probe_dst_prefix,
            groupArray((probe_ttl, cumulative_probes)) AS probes_per_ttl
        FROM {probes_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (probe_protocol, probe_dst_prefix)
        """


class GetProbesDiff(ProbesQuery):
    """
    Return the number of probes sent at a specific round and at the previous round.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetProbesDiff
        >>> row = GetProbesDiff(round_eq=1).execute(client, 'test_nsdi_example')[0]
        >>> row["probe_protocol"]
        1
        >>> row["probe_dst_prefix"]
        '::ffff:200.0.0.0'
        >>> sorted(row["probes_per_ttl"])
        [[1, 6, 0], [2, 6, 0], [3, 6, 0], [4, 6, 0]]
        >>> row = GetProbesDiff(round_eq=2).execute(client, 'test_nsdi_example')[0]
        >>> sorted(row["probes_per_ttl"])
        [[1, 11, 6], [2, 18, 6], [3, 18, 6], [4, 18, 6]]
        >>> row = GetProbesDiff(round_eq=3).execute(client, 'test_nsdi_example')[0]
        >>> sorted(row["probes_per_ttl"])
        [[1, 11, 11], [2, 20, 18], [3, 27, 18], [4, 27, 18]]
        >>> GetProbesDiff(round_eq=4).execute(client, 'test_nsdi_example')
        []
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq
        return f"""
        SELECT
            current.probe_protocol,
            current.probe_dst_prefix,
            groupArray((current.probe_ttl, current.cumulative_probes, previous.cumulative_probes)) AS probes_per_ttl
        FROM {probes_table(measurement_id)} AS current
        LEFT JOIN (
            SELECT
                probe_protocol,
                probe_dst_prefix,
                probe_ttl,
                cumulative_probes
            FROM {probes_table(measurement_id)}
            WHERE {ip_in("probe_dst_prefix", subset)} AND round = {self.round_eq - 1}
        ) AS previous
        ON current.probe_protocol = previous.probe_protocol
        AND current.probe_dst_prefix = previous.probe_dst_prefix
        AND current.probe_ttl = previous.probe_ttl
        WHERE {self.filters(subset)}
        GROUP BY (current.probe_protocol, current.probe_dst_prefix)
        """



"""
Wrappers around ClickHouse SQL queries.

The queries operate on different kind of tables.
Refer to the following superclasses for more information:
[Query][diamond_miner.queries.Query],
[LinksQuery][diamond_miner.queries.LinksQuery],
[PrefixesQuery][diamond_miner.queries.PrefixesQuery],
[ProbesQuery][diamond_miner.queries.ProbesQuery],
[ResultsQuery][diamond_miner.queries.ResultsQuery].
"""
from .count import Count
from .count_rows import CountLinksPerPrefix, CountProbesPerPrefix, CountResultsPerPrefix
from .create_links_table import CreateLinksTable
from .create_prefixes_table import CreatePrefixesTable
from .create_probes_table import CreateProbesTable
from .create_results_table import CreateResultsTable
from .create_tables import CreateTables
from .drop_tables import DropTables
from .get_invalid_prefixes import (
    GetInvalidPrefixes,
    GetPrefixesWithAmplification,
    GetPrefixesWithLoops,
)
from .get_links import GetLinks
from .get_links_from_results import GetLinksFromResults
from .get_mda_probes import GetMDAProbes
from .get_nodes import GetNodes
from .get_prefixes import GetPrefixes
from .get_probes import GetProbes, GetProbesDiff
from .get_results import GetResults
from .get_sliding_prefixes import GetSlidingPrefixes
from .insert_links import InsertLinks
from .insert_mda_probes import InsertMDAProbes
from .insert_prefixes import InsertPrefixes
from .insert_results import InsertResults
from .query import (
    LinksQuery,
    PrefixesQuery,
    ProbesQuery,
    Query,
    ResultsQuery,
    StoragePolicy,
    links_table,
    prefixes_table,
    probes_table,
    results_table,
)

__all__ = (
    "Count",
    "CountLinksPerPrefix",
    "CountProbesPerPrefix",
    "CountResultsPerPrefix",
    "CreateLinksTable",
    "CreatePrefixesTable",
    "CreateProbesTable",
    "CreateResultsTable",
    "CreateTables",
    "DropTables",
    "GetLinks",
    "GetLinksFromResults",
    "GetMDAProbes",
    "GetNodes",
    "GetPrefixes",
    "GetProbes",
    "GetProbesDiff",
    "GetResults",
    "GetSlidingPrefixes",
    "GetInvalidPrefixes",
    "GetPrefixesWithAmplification",
    "GetPrefixesWithLoops",
    "InsertMDAProbes",
    "InsertLinks",
    "InsertPrefixes",
    "InsertResults",
    "Query",
    "LinksQuery",
    "PrefixesQuery",
    "ProbesQuery",
    "ResultsQuery",
    "StoragePolicy",
    "links_table",
    "prefixes_table",
    "probes_table",
    "results_table",
)



from dataclasses import dataclass

from diamond_miner.defaults import DEFAULT_FAILURE_RATE, UNIVERSE_SUBSET
from diamond_miner.queries.query import LinksQuery, links_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetMDAProbes(LinksQuery):
    """
    Return the number of probes to send per prefix and per TTL according to the Diamond-Miner algorithm.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetMDAProbes
        >>> GetMDAProbes(round_leq=1).execute(client, "test_nsdi_lite")
        [{'probe_protocol': 1, 'probe_dst_prefix': '::ffff:200.0.0.0', 'cumulative_probes': [12, 12, 12, 12], 'TTLs': [1, 2, 3, 4]}]
    """

    adaptive_eps: bool = True

    dminer_lite: bool = True
    "If true, use an heuristic that requires less probes to handle nested load-balancers."

    target_epsilon: float = DEFAULT_FAILURE_RATE
    """
    The desired failure rate of the MDA algorithm, that is, the probability of not detecting
    all the outgoing edges of a load-balancer for a given prefix and TTL.
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.adaptive_eps:
            eps_fragment = """
            arrayMax(links_per_ttl) AS max_links,
            if(max_links == 0, target_epsilon, 1 - exp(log(1 - target_epsilon) / max_links))
                AS epsilon,
            """
        else:
            eps_fragment = """
            target_epsilon AS epsilon,
            """

        if self.dminer_lite:
            dm_fragment = """
            arrayMap(k -> toUInt32(ceil(ln(epsilon / (k + 1)) / ln((k + 1 - 1) / (k + 1)))), links_per_ttl) AS mda_flows,
            """
        else:
            # TODO: Implement by computing Dh(v)
            raise NotImplementedError

        return f"""
        WITH
            {self.target_epsilon} AS target_epsilon,
            -- 1) Compute the links
            --  x.1       x.2        x.3
            -- (near_ttl, near_addr, far_addr)
            groupUniqArray((near_ttl, near_addr, far_addr)) AS links,
            -- 2) Count the number of links per TTL
            -- extract only the TTLs, this greatly speeds-up arrayCount
            arrayMap(x -> x.1, links) AS links_ttls,
            -- find the min/max TTLs
            -- we add +2 since range() is exclusive and that we compute the max over the *near* TTL
            range(arrayMin(links_ttls), arrayMax(links_ttls) + 2) AS TTLs,
            -- count distinct links per TTL
            arrayMap(t -> countEqual(links_ttls, t), TTLs) AS links_per_ttl,
            -- 3) Compute MDA stopping points
            {eps_fragment}
            -- 4) Compute the number of probes to send during the next round
            {dm_fragment}
            -- compute the number of probes to send during the next round
            -- => max of probes to send over TTL t and t-1
            arrayMap(i -> arrayMax([mda_flows[i], mda_flows[i - 1]]), arrayEnumerate(TTLs)) AS cumulative_probes
            -- TODO: Cleanup/optimize/rewrite/... below
            -- do not send probes to TTLs where no replies have been received
            -- it is unlikely that we will discover more at this TTL if the first 6 flows have seen nothing
            -- (see GetNextRoundStateless)
        SELECT
            probe_protocol,
            probe_dst_prefix,
            cumulative_probes,
            TTLs
        FROM {links_table(measurement_id)} AS links_table
        WHERE {self.filters(subset)}
        GROUP BY (probe_protocol, probe_src_addr, probe_dst_prefix)
        """



from math import ceil, log

from diamond_miner.defaults import DEFAULT_FAILURE_RATE


def stopping_point(k: int, eps: float = DEFAULT_FAILURE_RATE) -> int:
    """
    Return the number `n_k` of probes that guarantees that the probability of not
    detecting `k` outgoing load-balanced edges is lower than `eps`[@veitch2009failure;@jacquet2018collecter].

    Examples:
        >>> stopping_point(1, 0.05)
        0
        >>> stopping_point(2, 0.05)
        6
        >>> stopping_point(3, 0.05)
        11
        >>> stopping_point(11, 0.05)
        57
        >>> stopping_point(101, 0.05)
        765

    Note:
        There is a typo in the D-Miner paper: n(101) = 765, not 757.
    """
    assert (k >= 1) and (0 <= eps <= 1)
    if k == 1:
        return 0
    return ceil(log(eps / k) / log((k - 1) / k))



import random
import shutil
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from tempfile import TemporaryDirectory

from pych_client import ClickHouseClient
from zstandard import ZstdCompressor

from diamond_miner.defaults import (
    DEFAULT_PREFIX_SIZE_V4,
    DEFAULT_PREFIX_SIZE_V6,
    DEFAULT_PROBE_DST_PORT,
    DEFAULT_PROBE_SRC_PORT,
)
from diamond_miner.format import format_probe
from diamond_miner.generators.database import probe_generator_from_database
from diamond_miner.logger import logger
from diamond_miner.mappers import SequentialFlowMapper
from diamond_miner.queries import GetProbesDiff
from diamond_miner.subsets import subsets_for
from diamond_miner.typing import FlowMapper, IPNetwork, Probe
from diamond_miner.utilities import available_cpus


def probe_generator_parallel(
    filepath: Path,
    client: ClickHouseClient,
    measurement_id: str,
    round_: int,
    *,
    mapper_v4: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V4),
    mapper_v6: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V6),
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    probe_ttl_geq: int | None = None,
    probe_ttl_leq: int | None = None,
    max_open_files: int = 8192,
    n_workers: int = max(available_cpus() // 8, 1),
) -> int:
    """
    Compute the probes to send given the previously discovered links.
    This function shuffle the probes on-disk:
    [External-memory shuffling in linear time?](https://lemire.me/blog/2010/03/15/external-memory-shuffling-in-linear-time/)

    Args:
        filepath: Output file (Zstd-compressed CSV file); will be overwritten.
        client: ClickHouse client.
        measurement_id: Measurement id.
        round_: Number of the round for which to generate the probes.
        mapper_v4: The flow mapper for IPv4 probes.
        mapper_v6: The flow mapper for IPv6 probes.
        probe_src_port: The minimum source port of the probes (can be incremented by the flow mapper).
        probe_dst_port: The destination port of the probes (constant).

    """
    # TODO: These subsets are sub-optimal, `CountProbesPerPrefix` should count
    # the actual number of probes to be sent, not the total number of probes sent.
    subsets = subsets_for(
        GetProbesDiff(
            round_eq=round_, probe_ttl_geq=probe_ttl_geq, probe_ttl_leq=probe_ttl_leq
        ),
        client,
        measurement_id,
    )

    if not subsets:
        return 0

    n_files_per_subset = max_open_files // len(subsets)

    logger.info(
        "mda_probes n_workers=%s n_subsets=%s n_files_per_subset=%s",
        n_workers,
        len(subsets),
        n_files_per_subset,
    )

    with TemporaryDirectory(dir=filepath.parent) as temp_dir:
        with ProcessPoolExecutor(n_workers) as executor:
            futures = [
                executor.submit(
                    worker,
                    Path(temp_dir) / f"subset_{i}",
                    client.config,
                    measurement_id,
                    round_,
                    mapper_v4,
                    mapper_v6,
                    probe_src_port,
                    probe_dst_port,
                    probe_ttl_geq,
                    probe_ttl_leq,
                    subset,
                    n_files_per_subset,
                )
                for i, subset in enumerate(subsets)
            ]
            n_probes = sum(future.result() for future in as_completed(futures))

        files = list(Path(temp_dir).glob("subset_*.csv.zst"))
        random.shuffle(files)

        logger.info("mda_probes status=merging n_files=%s", len(files))
        with filepath.open("wb") as out:
            for f in files:
                with f.open("rb") as inp:
                    shutil.copyfileobj(inp, out)

    return n_probes


def worker(
    prefix: Path,
    client_config: dict,
    measurement_id: str,
    round_: int,
    mapper_v4: FlowMapper,
    mapper_v6: FlowMapper,
    probe_src_port: int,
    probe_dst_port: int,
    probe_ttl_geq: int | None,
    probe_ttl_leq: int | None,
    subset: IPNetwork,
    n_files: int,
) -> int:
    """
    Execute the :class:`diamond_miner.queries.GetNextRound` query
    on the specified subset, and write the probes to the specified file.
    """
    # TODO: random.shuffle is slow...
    # A potentially simpler and better way would be to shuffle
    # the result of the next round query, which is small (~10M rows),
    # and to directly write the probes to a file depending on hash(flow_id).
    # e.g. ORDER BY rand() in GetNextRound.

    # The larger `max_probes_in_memory`, the better the performance
    # (less calls to `probes.clear()`)
    # and the randomization but the more the memory usage.
    max_probes_in_memory = 1_000_000

    outputs: list[tuple] = []
    for i in range(n_files):
        ctx = ZstdCompressor(level=1)
        file = prefix.with_suffix(f".{i}.csv.zst").open("wb")
        stream = ctx.stream_writer(file)
        outputs.append((ctx, file, stream))

    probes_by_file: list[list[Probe]] = [[] for _ in range(n_files)]
    n_probes = 0

    for probe in probe_generator_from_database(
        client=ClickHouseClient(**client_config),
        measurement_id=measurement_id,
        round_=round_,
        mapper_v4=mapper_v4,
        mapper_v6=mapper_v6,
        probe_src_port=probe_src_port,
        probe_dst_port=probe_dst_port,
        probe_ttl_geq=probe_ttl_geq,
        probe_ttl_leq=probe_ttl_leq,
        subsets=(subset,),
    ):
        # probe[:-2] => (probe_dst_addr, probe_src_port, probe_dst_port)
        probes_by_file[hash(probe[:-2]) % n_files].append(probe)
        n_probes += 1
        if n_probes % max_probes_in_memory == 0:
            flush(probes_by_file, outputs)

    flush(probes_by_file, outputs)

    for ctx, file, stream in outputs:
        stream.close()
        file.close()

    return n_probes


def flush(probes_by_file: list[list[Probe]], outputs: list[tuple]) -> None:
    for file_id, probes in enumerate(probes_by_file):
        _, _, stream = outputs[file_id]
        random.shuffle(probes)
        for probe in probes:
            stream.write(format_probe(*probe).encode("ascii") + b"\n")
        probes.clear()



from ipaddress import IPv6Network

from pych_client import ClickHouseClient

from diamond_miner.queries import (
    CountLinksPerPrefix,
    CountProbesPerPrefix,
    CountResultsPerPrefix,
    LinksQuery,
    ProbesQuery,
    ResultsQuery,
)
from diamond_miner.utilities import common_parameters

ALL_ONES_V6 = (2**128) - 1
Counts = dict[IPv6Network, int]


def subsets_for(
    query: LinksQuery | ProbesQuery | ResultsQuery,
    client: ClickHouseClient,
    measurement_id: str,
    *,
    max_items_per_subset: int = 8_000_000,
) -> list[IPv6Network]:
    """
    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetLinks, GetProbes, GetResults
        >>> subsets_for(GetLinks(), client, 'test_nsdi_example', max_items_per_subset=1)
        [IPv6Network('::ffff:c800:0/112')]
        >>> subsets_for(GetProbes(round_eq=1), client, 'test_nsdi_example', max_items_per_subset=1)
        [IPv6Network('::ffff:c800:0/112')]
        >>> subsets_for(GetResults(), client, 'test_nsdi_example', max_items_per_subset=1)
        [IPv6Network('::ffff:c800:0/112')]
    """
    if isinstance(query, LinksQuery):
        count_query = CountLinksPerPrefix(**common_parameters(query, LinksQuery))
    elif isinstance(query, ProbesQuery):
        count_query = CountProbesPerPrefix(**common_parameters(query, ProbesQuery))  # type: ignore
    elif isinstance(query, ResultsQuery):
        count_query = CountResultsPerPrefix(**common_parameters(query, ResultsQuery))  # type: ignore
    else:
        raise NotImplementedError
    counts = {
        addr_to_network(
            row["prefix"], count_query.prefix_len_v4, count_query.prefix_len_v6
        ): row["count"]
        for row in count_query.execute_iter(client, measurement_id)
    }
    return split(counts, max_items_per_subset)


def split(counts: Counts, max_items_per_subset: int) -> list[IPv6Network]:
    """
    Return the IP networks such that there are no more than `max_items_per_subset`
    per network.

    Args:
        counts: Number of items per prefix in the database table.
        max_items_per_subset: Maximum number of items per network.

    Examples:
        >>> counts = {IPv6Network("::ffff:8.8.4.0/120"): 10, IPv6Network("::ffff:8.8.8.0/120"): 5}
        >>> split(counts, 15)
        [IPv6Network('::/0')]
        >>> split(counts, 10)
        [IPv6Network('::ffff:808:0/117'), IPv6Network('::ffff:808:800/117')]
        >>> split(counts, 1) # Impossible case, should return the minimal feasible networks.
        [IPv6Network('::ffff:808:400/120'), IPv6Network('::ffff:808:800/120')]
        >>> split({}, 10)
        []
    """
    candidates = [(IPv6Network("::/0"), n_items(counts, IPv6Network("::/0")))]
    subsets = []

    while candidates:
        candidate, n_replies = candidates.pop()
        if max_items_per_subset >= n_replies > 0:
            subsets.append(candidate)
        elif n_replies > 0:
            a, b = tuple(candidate.subnets(prefixlen_diff=1))
            n_items_a = n_items(counts, a)
            n_items_b = n_items(counts, b)
            if n_items_a + n_items_b == 0:
                subsets.append(candidate)
            else:
                candidates.append((a, n_items_a))
                candidates.append((b, n_items_b))

    return sorted(subsets)


def addr_to_network(addr: str, prefix_len_v4: int, prefix_len_v6: int) -> IPv6Network:
    """
    Examples:
        >>> addr_to_network("::ffff:8.8.8.0", 24, 64)
        IPv6Network('::ffff:808:800/120')
        >>> addr_to_network("2001:4860:4860:1234::", 24, 64)
        IPv6Network('2001:4860:4860:1234::/64')
    """
    assert ":" in addr, "`addr` must be an (IPv4-mapped) IPv6 address."
    if addr.startswith("::ffff:"):
        return IPv6Network(f"{addr}/{96+prefix_len_v4}")
    return IPv6Network(f"{addr}/{prefix_len_v6}")


def n_items(counts: Counts, subset: IPv6Network) -> int:
    """
    Examples:
        >>> counts = {IPv6Network("1000::/16"): 2, IPv6Network("8000::/16"): 10}
        >>> n_items(counts, IPv6Network("0000::/1"))
        2
        >>> n_items(counts, IPv6Network("8000::/1"))
        10
        >>> n_items(counts, IPv6Network("::/0"))
        12
    """
    total = 0
    for network, count in counts.items():
        if is_subnet_of(network, subset):
            total += count
    return total


def is_subnet_of(a: IPv6Network, b: IPv6Network) -> bool:
    """
    A faster version of IPv6Network.subnet_of(other).

    Examples:
        >>> is_subnet_of(IPv6Network("1000::/16"), IPv6Network("1000::/16"))
        True
        >>> is_subnet_of(IPv6Network("1000::/17"), IPv6Network("1000::/16"))
        True
        >>> is_subnet_of(IPv6Network("1000::/15"), IPv6Network("1000::/16"))
        False
        >>> is_subnet_of(IPv6Network("1000::/16"), IPv6Network("2000::/16"))
        False
    """
    a_net = a.network_address._ip  # type: ignore
    b_net = b.network_address._ip  # type: ignore
    if b_net <= a_net:
        a_brd = a_net | (a.netmask._ip ^ ALL_ONES_V6)  # type: ignore
        b_brd = b_net | (b.netmask._ip ^ ALL_ONES_V6)  # type: ignore
        return b_brd >= a_brd  # type: ignore
    return False



from collections.abc import Iterable, Iterator, Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import datetime
from functools import reduce
from typing import Any

from pych_client import ClickHouseClient

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.logger import logger
from diamond_miner.queries.fragments import (
    and_,
    eq,
    geq,
    ip_eq,
    ip_in,
    leq,
    lt,
    not_,
    or_,
)
from diamond_miner.typing import IPNetwork
from diamond_miner.utilities import LoggingTimer, available_cpus


def links_table(measurement_id: str) -> str:
    """Returns the name of the links table."""
    return f"links__{measurement_id}".replace("-", "_")


def prefixes_table(measurement_id: str) -> str:
    """Returns the name of the prefixes table."""
    return f"prefixes__{measurement_id}".replace("-", "_")


def probes_table(measurement_id: str) -> str:
    """Returns the name of the probes table."""
    return f"probes__{measurement_id}".replace("-", "_")


def results_table(measurement_id: str) -> str:
    """Returns the name of the results table."""
    return f"results__{measurement_id}".replace("-", "_")


@dataclass(frozen=True)
class StoragePolicy:
    """
    - [TTL for Columns and Tables](https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-ttl)
    - [Using Multiple Block Devices for Data Storage](https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-multiple-volumes)
    """

    name: str = "default"
    """Name of the ClickHouse storage policy to use for the table."""
    archive_to: str = "default"
    """Name of the ClickHouse archive volume."""
    archive_on: datetime = datetime(2100, 1, 1)
    """Date at which the table will be moved to the archive volume."""


@dataclass(frozen=True)
class Query:
    """Base class for every query."""

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # As a query user, prefer calling `statements` instead of `statement` as there
        # is no guarantees that the query will implement this method and return a single statement.
        raise NotImplementedError

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        # Override this method if you want your query to return multiple statements.
        return (self.statement(measurement_id, subset),)

    def execute(
        self,
        client: ClickHouseClient,
        measurement_id: str,
        *,
        data: Any | None = None,
        limit: tuple[int, int] | None = None,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
    ) -> list[dict]:
        """
        Execute the query and return each row as a dict.
        Args:
            client: ClickHouse client.
            measurement_id: Measurement id.
            data: str or bytes iterator containing data to send.
            limit: (limit, offset) tuple.
            subsets: Iterable of IP networks on which to execute the query independently.
        """
        rows = []
        for subset in subsets:
            for i, statement in enumerate(self.statements(measurement_id, subset)):
                with LoggingTimer(
                    logger,
                    f"query={self.name}#{i} measurement_id={measurement_id} subset={subset} limit={limit}",
                ):
                    settings = dict(
                        limit=limit[0] if limit else 0,
                        offset=limit[1] if limit else 0,
                    )
                    rows += client.json(statement, data=data, settings=settings)
        return rows

    def execute_iter(
        self,
        client: ClickHouseClient,
        measurement_id: str,
        *,
        data: Any | None = None,
        limit: tuple[int, int] | None = None,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
    ) -> Iterator[dict]:
        """
        Execute the query and return each row as a dict, as they are received from the database.
        """
        for subset in subsets:
            for i, statement in enumerate(self.statements(measurement_id, subset)):
                with LoggingTimer(
                    logger,
                    f"query={self.name}#{i} measurement_id={measurement_id} subset={subset} limit={limit}",
                ):
                    settings = dict(
                        limit=limit[0] if limit else 0,
                        offset=limit[1] if limit else 0,
                    )
                    yield from client.iter_json(statement, data=data, settings=settings)

    def execute_concurrent(
        self,
        client: ClickHouseClient,
        measurement_id: str,
        *,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: tuple[int, int] | None = None,
        concurrent_requests: int = max(available_cpus() // 8, 1),
    ) -> None:
        """
        Execute the query concurrently on the specified subsets.
        """
        logger.info("query=%s concurrent_requests=%s", self.name, concurrent_requests)
        with ThreadPoolExecutor(concurrent_requests) as executor:
            futures = [
                executor.submit(
                    self.execute,
                    client=client,
                    measurement_id=measurement_id,
                    subsets=(subset,),
                    limit=limit,
                )
                for subset in subsets
            ]
            for future in as_completed(futures):
                future.result()


@dataclass(frozen=True)
class LinksQuery(Query):
    """Base class for queries on the links table."""

    filter_inter_round: bool = False
    "If true, exclude links inferred across rounds."

    filter_partial: bool = False
    "If true, exclude partial links: `('::', node)` and `(node, '::')`."

    filter_virtual: bool = False
    "If true, exclude virtual links: `('::', '::')`."

    near_or_far_addr: str | None = None
    "If specified, keep only the links that contains this IP address."

    probe_protocol: int | None = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: str | None = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: int | None = None
    "If specified, keep only the links from this round."

    round_leq: int | None = None
    "If specified, keep only the links from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """`WHERE` clause common to all queries on the links table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("near_round", self.round_eq), eq("far_round", self.round_eq)]
        if self.round_leq:
            s += [leq("near_round", self.round_leq), leq("far_round", self.round_leq)]
        if self.near_or_far_addr:
            s += [
                or_(
                    ip_eq("near_addr", self.near_or_far_addr),
                    ip_eq("far_addr", self.near_or_far_addr),
                )
            ]
        if self.filter_inter_round:
            s += [not_("is_inter_round")]
        if self.filter_partial:
            s += [not_("is_partial")]
        if self.filter_virtual:
            s += [not_("is_virtual")]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class PrefixesQuery(Query):
    """Base class for queries on the prefixes table."""

    probe_protocol: int | None = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: str | None = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    def filters(self, subset: IPNetwork) -> str:
        """`WHERE` clause common to all queries on the prefixes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ProbesQuery(Query):
    """Base class for queries on the probes table."""

    probe_protocol: int | None = None
    "If specified, keep only probes sent with this protocol."

    probe_ttl_geq: int | None = None
    "If specified, keep only the probes with TTL >= this value."

    probe_ttl_leq: int | None = None
    "If specified, keep only the probes with TTL <= this value."

    round_eq: int | None = None
    "If specified, keep only the probes from this round."

    round_geq: int | None = None
    "If specified, keep only the probes from this round or after."

    round_leq: int | None = None
    "If specified, keep only the probes from this round or before."

    round_lt: int | None = None
    "If specified, keep only the probes from before this round."

    def filters(self, subset: IPNetwork) -> str:
        """`WHERE` clause common to all queries on the probes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_ttl_geq:
            s += [geq("probe_ttl", self.probe_ttl_geq)]
        if self.probe_ttl_leq:
            s += [leq("probe_ttl", self.probe_ttl_leq)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_geq:
            s += [geq("round", self.round_geq)]
        if self.round_lt:
            s += [lt("round", self.round_lt)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ResultsQuery(Query):
    """Base class for queries on the results table."""

    filter_destination_host: bool = True
    "If true, ignore the replies from the destination host."

    filter_destination_prefix: bool = True
    "If true, ignore the replies from the destination prefix."

    filter_private: bool = True
    "If true, ignore the replies from private IP addresses."

    filter_invalid_probe_protocol: bool = True
    "If true, ignore the replies with probe protocol ≠ ICMP, ICMPv6 or UDP."

    time_exceeded_only: bool = True
    "If true, ignore non ICMP time exceeded replies."

    probe_protocol: int | None = None
    "If specified, keep only the replies to probes sent with this protocol."

    probe_src_addr: str | None = None
    """
    If specified, keep only the replies to probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: int | None = None
    "If specified, keep only the replies from this round."

    round_leq: int | None = None
    "If specified, keep only the replies from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """`WHERE` clause common to all queries on the results table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        if self.filter_destination_host:
            s += [not_("destination_host_reply")]
        if self.filter_destination_prefix:
            s += [not_("destination_prefix_reply")]
        if self.filter_private:
            s += [not_("private_probe_dst_prefix"), not_("private_reply_src_addr")]
        if self.time_exceeded_only:
            s += ["time_exceeded_reply"]
        if self.filter_invalid_probe_protocol:
            s += ["valid_probe_protocol"]
        return reduce(and_, s or ["1"])



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import Query, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class InsertResults(Query):
    """
    Insert measurement results from a CSV file.
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"INSERT INTO {results_table(measurement_id)} FORMAT CSVWithNames"



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import cut_ipv6
from diamond_miner.queries.query import (
    LinksQuery,
    ProbesQuery,
    ResultsQuery,
    links_table,
    probes_table,
    results_table,
)
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CountLinksPerPrefix(LinksQuery):
    """
    Count the number of (non-distinct) links per prefix.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import CountLinksPerPrefix
        >>> rows = CountLinksPerPrefix().execute(client, 'test_nsdi_example')
        >>> sorted((row["prefix"], row["count"]) for row in rows)
        [('::ffff:200.0.0.0', 58)]
    """

    prefix_len_v4: int = 16
    "The IPv4 prefix length to consider."

    prefix_len_v6: int = 8
    "The IPv6 prefix length to consider."

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        WITH {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, count() AS count
        FROM {links_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """


@dataclass(frozen=True)
class CountProbesPerPrefix(ProbesQuery):
    """
    Count the number of probes sent per prefix.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import CountProbesPerPrefix
        >>> rows = CountResultsPerPrefix(round_eq=1).execute(client, 'test_nsdi_example')
        >>> sorted((row["prefix"], row["count"]) for row in rows)
        [('::ffff:200.0.0.0', 24)]
    """

    prefix_len_v4: int = 16
    "The IPv4 prefix length to consider."

    prefix_len_v6: int = 8
    "The IPv6 prefix length to consider."

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq
        return f"""
        WITH {cut_ipv6('probe_dst_prefix', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, sum(cumulative_probes) AS count
        FROM {probes_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """


@dataclass(frozen=True)
class CountResultsPerPrefix(ResultsQuery):
    """
    Count the number of results per prefix.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import CountResultsPerPrefix
        >>> rows = CountResultsPerPrefix(prefix_len_v4=8, prefix_len_v6=8).execute(client, 'test_count_replies')
        >>> sorted((row["prefix"], row["count"]) for row in rows)
        [('::ffff:1.0.0.0', 2), ('::ffff:2.0.0.0', 1), ('::ffff:204.0.0.0', 1)]
    """

    prefix_len_v4: int = 16
    "The IPv4 prefix length to consider."

    prefix_len_v6: int = 8
    "The IPv6 prefix length to consider."

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        WITH {cut_ipv6('probe_dst_addr', self.prefix_len_v4, self.prefix_len_v6)} AS prefix
        SELECT prefix, count() AS count
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY prefix
        """



from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from pych_client import ClickHouseClient

from diamond_miner.defaults import (
    DEFAULT_FAILURE_RATE,
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    PROTOCOLS,
    UNIVERSE_SUBSET,
)
from diamond_miner.format import format_ipv6
from diamond_miner.generators.standalone import split_prefix
from diamond_miner.queries.insert_mda_probes import InsertMDAProbes
from diamond_miner.queries.query import Query, probes_table
from diamond_miner.subsets import subsets_for
from diamond_miner.typing import IPNetwork
from diamond_miner.utilities import available_cpus


@dataclass(frozen=True)
class InsertProbes(Query):
    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"INSERT INTO {probes_table(measurement_id)} FORMAT JSONCompactEachRow"


def insert_probe_counts(
    client: ClickHouseClient,
    measurement_id: str,
    round_: int,
    prefixes: Iterable[tuple[str, str, Iterable[int], int]],
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
) -> None:
    """
    Insert the probe counts specified by `prefixes` into the probes table.

    Args:
        client: ClickHouse client.
        measurement_id: Measurement id.
        round_: Round number for which to insert the probe counts.
        prefixes: A list of `(prefix, protocol, ttls, n_probes)` tuples. /32 or /128 is assumed if not specified.
        prefix_len_v4: The prefix length to which the IPv4 prefixes will be split to.
        prefix_len_v6: The prefix length to which the IPv6 prefixes will be split to.

    Examples:
        >>> from diamond_miner.test import client, create_tables
        >>> from diamond_miner.queries import GetProbes
        >>> create_tables(client, "test_probe_counts")
        >>> insert_probe_counts(client, "test_probe_counts", 1, [("8.8.0.0/22", "icmp", range(2, 5), 6)])
        >>> rows = sorted(GetProbes(round_eq=1).execute(client, "test_probe_counts"), key=lambda x: x["probe_dst_prefix"])
        >>> len(rows)
        4
        >>> row = rows[0]
        >>> row["probe_dst_prefix"]
        '::ffff:8.8.0.0'
        >>> sorted(row["probes_per_ttl"])
        [[2, 6], [3, 6], [4, 6]]
        >>> row = rows[1]
        >>> row["probe_dst_prefix"]
        '::ffff:8.8.1.0'
        >>> sorted(row["probes_per_ttl"])
        [[2, 6], [3, 6], [4, 6]]
    """

    def gen() -> Iterator[bytes]:
        for prefix, protocol, ttls, n_probes in prefixes:
            protocol = PROTOCOLS[protocol]  # type: ignore
            for af, subprefix, subprefix_size in split_prefix(
                prefix, prefix_len_v4, prefix_len_v6
            ):
                yield "\n".join(
                    f'[{protocol},"{format_ipv6(subprefix)}",{ttl},{n_probes},{round_}]'
                    for ttl in ttls
                ).encode()

    InsertProbes().execute(client, measurement_id, data=gen())


def insert_mda_probe_counts(
    client: ClickHouseClient,
    measurement_id: str,
    previous_round: int,
    adaptive_eps: bool = False,
    target_epsilon: float = DEFAULT_FAILURE_RATE,
    concurrent_requests: int = max(available_cpus() // 8, 1),
) -> None:
    """
    Run the Diamond-Miner algorithm and insert the resulting probes into the probes table.

    Args:
        client: ClickHouse client.
        measurement_id: Measurement id.
        previous_round: Round on which to run the Diamond-Miner algorithm.
        adaptive_eps: Set to `True` to handle nested load-balancers.
        target_epsilon: Target failure rate of the MDA algorithm.
        concurrent_requests: Maximum number of requests to execute concurrently.
    """
    # TODO: set filter_partial and filter_virtual to false?
    query = InsertMDAProbes(
        adaptive_eps=adaptive_eps,
        round_leq=previous_round,
        filter_partial=True,
        filter_virtual=True,
        filter_inter_round=True,
        target_epsilon=target_epsilon,
    )
    subsets = subsets_for(query, client, measurement_id)
    query.execute_concurrent(
        client, measurement_id, subsets=subsets, concurrent_requests=concurrent_requests
    )



from os import environ

from pych_client import ClickHouseClient

# TODO: Rename to base_url
url = environ.get("DIAMOND_MINER_TEST_DATABASE_URL", "http://localhost:8123")

base_url = environ.get("DIAMOND_MINER_TEST_DATABASE_URL", "http://localhost:8123")
client = ClickHouseClient(base_url=base_url, database="default", username="default")


def create_tables(client: ClickHouseClient, measurement_id: str) -> None:
    # Avoid circular imports.
    from diamond_miner.queries.create_tables import CreateTables
    from diamond_miner.queries.drop_tables import DropTables

    DropTables().execute(client, measurement_id)
    CreateTables().execute(client, measurement_id)



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.get_mda_probes import GetMDAProbes
from diamond_miner.queries.query import probes_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class InsertMDAProbes(GetMDAProbes):
    """
    Insert the result of the `GetMDAProbes` queries
    into the probes table.
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_leq
        return f"""
        INSERT INTO {probes_table(measurement_id)}
        WITH
            arrayJoin(arrayZip(TTLs, cumulative_probes)) AS ttl_probe
        SELECT
            probe_protocol,
            probe_dst_prefix,
            ttl_probe.1,
            ttl_probe.2,
            {self.round_leq + 1}
        FROM ({super().statement(measurement_id, subset)})
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import Query
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class Count(Query):
    """
    Count the number of rows returned by a given query.

    Examples:
        >>> from diamond_miner.test import client
        >>> from diamond_miner.queries import GetLinks, GetNodes
        >>> Count(query=GetNodes()).execute(client, 'test_nsdi_example')[0]["count()"]
        7
        >>> Count(query=GetLinks()).execute(client, 'test_nsdi_example')[0]["count()"]
        8
    """

    query: Query | None = None
    "The query for which to count the nodes."

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # `query` must be typed `Optional` since it appears after arguments with default values.
        assert self.query is not None
        return f"SELECT count() FROM ({self.query.statement(measurement_id, subset)})"



from collections.abc import Iterator, Sequence
from typing import Any

class ParameterGrid:
    def __init__(self, *parameters: Sequence): ...
    def __getitem__(self, index: Sequence[int] | int) -> Sequence[Any]: ...
    def __iter__(self) -> Iterator[Sequence[Any]]: ...
    def __len__(self) -> int: ...
    @property
    def size(self) -> int: ...
    def shuffled(
        self, rounds: int = ..., seed: int | None = ...
    ) -> Iterator[Sequence[Any]]: ...
